@@ -14,6 +14,25 @@ import urllib3
 urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
 
 
+def get_exlist(ex) :
+    setex = list()
+    if os.path.isfile(ex) :
+        for ln in open(ex,"r").readlines() :
+            setex.append(ln)
+    else :
+        for e in ex.split(",") :
+            setex.append(e)
+    return setex
+
+def get_encoding(txt) :
+    if not txt :
+        return None
+    # <meta HTTP-EQUIV="Content-Type" content="text/html; charset=gb2312" />
+    m = re.search(r"<meta\s+.*?\s+charset=(\w+)",txt)
+    if m :
+        return m.group(1)
+    else :
+        return None
 def build_index(dirbook) :
     from time import strftime,localtime
     if not os.path.isdir(dirbook) :
@@ -41,7 +60,7 @@ def build_index(dirbook) :
         if not os.path.exists(fidx) :
             continue
         ic += 1
-        with open(fidx,"r") as f :
+        with open(fidx,"r",encoding="UTF-8") as f :
             cnt = 0
             for ln in reversed(f.readlines()) :
                 if cnt < 2 and re.search(r"td.*href.*html",ln) :
@@ -59,11 +78,11 @@ def build_index(dirbook) :
                     cnt += 1
     content +=("</table>\n")
     content +=("</html>\n")
-    with open(os.path.join(dirbook,"index.html"),"w") as fw  :
+    with open(os.path.join(dirbook,"index.html"),"w",encoding="UTF-8") as fw  :
         fw.write(content)
     
 
-def adhoc_fetch(title,page,limit=10,tgtdir=os.getcwd(),debug=False) :
+def adhoc_fetch(title,page,limit=10,tgtdir=os.getcwd(),debug=False,force=0,root=None,exlist=None) :
     def random_useragent() :
         ntvers = ["9.0","10.0","8.0"]
         os = [ "Windows NT "+str(v) for v in ntvers ] + ["Macintosh"] * 3
@@ -78,23 +97,28 @@ def adhoc_fetch(title,page,limit=10,tgtdir=os.getcwd(),debug=False) :
         print("# tgtdir {} is missing.".format(tgtdir),file=sys.stderr,flush=True)
     if not os.path.isdir(os.path.join(tgtdir,title)) :
         os.mkdir(os.path.join(tgtdir,title))
-        with open(os.path.join(tgtdir,title,"index.html"),"w") as f :
+        with open(os.path.join(tgtdir,title,"index.html"),"w",encoding="UTF-8") as f :
             header = """
 <meta name="viewport" content="width=device-width, initial-scale=1.0">
+<meta http-equiv="Content-Type" content="text/html; charset=UTF-8"/>
 <title>book_title</title>
 <body style="background-color: #212121;color: #B5B5B5;">
 <table style="border-collapse:collapse">
 """
             f.write(header.replace("book_title",title))
-        with open(os.path.join(tgtdir,title, "downloaded.txt"),"w") as f :
+        with open(os.path.join(tgtdir,title, "downloaded.txt"),"w",encoding="UTF-8") as f :
             pass
     downloded = set()
     last = ""
     if os.path.isfile(os.path.join(tgtdir,title, "downloaded.txt")) :
-        with open(os.path.join(tgtdir,title, "downloaded.txt"),"r") as f :
+        with open(os.path.join(tgtdir,title, "downloaded.txt"),"r",encoding="UTF-8") as f :
             downloaded = set([ln.rstrip() for ln in f.readlines()])
-            if downloaded :
+            if force and len(downloaded) > 1 :
+                last = sorted(list(downloaded))[-(force+1)]
+            elif downloaded :
                 last = max(downloaded)
+            else :
+                pass
     try :
         user_agent = {'User-Agent': random_useragent()}
         if debug :
@@ -104,12 +128,18 @@ def adhoc_fetch(title,page,limit=10,tgtdir=os.getcwd(),debug=False) :
     except :
         print("# {}".format(traceback.format_exc().splitlines()[-1]),file=sys.stderr,flush=True)
         return "# No update."
-    rsp.encoding = rsp.apparent_encoding
+    xcode = get_encoding(rsp.text) 
+    if xcode :
+        rsp.encoding = xcode.upper()
+    else :
+        rsp.encoding = rsp.apparent_encoding
     rtext = rsp.text
     if debug :
         print(rtext)
     chapters = list()
-    for m in re.findall(r"(<td|<dd|<li)[^<>]*?>\s*<a\s+href=\"(\S*?\d+\.(htm|html))\".*?>\s*(.*?)\s*</a>",rtext,re.DOTALL) :
+    for m in re.findall(r"(<td|<dd|<li|<span)[^<>]*?>\s*<a\s+href\s*=\s*\"(\S*?\d+\.(htm|html))\".*?>\s*(.*?)\s*</a>",rtext,re.DOTALL) :
+        if debug :
+            print(m)
         url=m[1].split("/")[-1]
         m2 = re.search(r"(\d+)",url)
         if m2 :
@@ -118,6 +148,8 @@ def adhoc_fetch(title,page,limit=10,tgtdir=os.getcwd(),debug=False) :
         else :
             continue
         chapter=m[3]
+        chapter = re.sub(r"（.*）","",chapter)
+        chapter = re.sub(r"\(.*\)","",chapter)
         chapters.append((url,chapter))
     def cid(c) :
         s = c[0]
@@ -129,8 +161,10 @@ def adhoc_fetch(title,page,limit=10,tgtdir=os.getcwd(),debug=False) :
             return 2**(30+len(m.group(1)))+int(m.group(1))
         else :
             return 2**60
+    maxc = "n/a"
     chapters = sorted(list(set(chapters)),key=cid)
-    maxc = chapters[-1][0]
+    if chapters :
+        maxc = chapters[-1][0]
     if maxc == last :
         print("# No update.",file=sys.stderr,flush=True)
         return "# No update : {}".format(title)
@@ -147,7 +181,10 @@ def adhoc_fetch(title,page,limit=10,tgtdir=os.getcwd(),debug=False) :
         if cname and cname < last :
             continue
         if page.endswith(".htm") or page.endswith(".html") :
-            page = "".join(page.split("/")[:-1])
+            if root :
+                page = root
+            else :
+                page = "".join(page.split("/")[:-1])
         if not url.startswith("http") :
             url = page.rstrip("/") + "/" + url
         chapter=m[1]
@@ -158,7 +195,11 @@ def adhoc_fetch(title,page,limit=10,tgtdir=os.getcwd(),debug=False) :
                 print("# {}".format(url),file=sys.stderr,flush=True)
                 print("# {}".format(user_agent),file=sys.stderr,flush=True)
             rsp = ss.get(url,headers=user_agent,timeout=20,verify=False)
-            rsp.encoding = rsp.apparent_encoding
+            xcode = get_encoding(rsp.text) 
+            if xcode :
+                rsp.encoding = xcode.upper()
+            else :
+                rsp.encoding = rsp.apparent_encoding
         except :
            print("# {}".format(traceback.format_exc().splitlines()[-1]),file=sys.stderr,flush=True)
            break
@@ -188,13 +229,28 @@ def adhoc_fetch(title,page,limit=10,tgtdir=os.getcwd(),debug=False) :
                 if started :
                     context += ln
         if len(context) < 200 :
-            print("# context length abnormal. most likely hit with rate limiter", file=sys.stderr, flush=True)
-            break
-        context = re.sub(r"</br>","<br/>\n",context)
-        context = re.sub(r"<br\s*/>","<br/>\n",context)
-        if not re.search(r"<br/>\s*<br/>",context,re.DOTALL) :
-            context = context.replace("<br/>","<br/><br/>")
-        context = context.replace("<br/>","<br/>\n")
+            print("# context length abnormal. chapter skipped.", file=sys.stderr, flush=True)
+            continue
+        context = re.sub(r"</br>","<br/>",context)
+        context = re.sub(r"<br\s+/>","<br/>",context)
+        context = context.replace("<br/>","<br/><br/>")
+        if re.search(r"<br/>\s*<br/>\s*<br/>",context,flags=re.DOTALL) :
+            context = re.sub(r"<br/>\s*<br/>\s*<br/>","<br/>",context,flags=re.DOTALL)
+        context = context.replace("<br/>","\n<br/>\n")
+        context = context.replace("&nbsp;","")
+        context = re.sub(r"\<div\>.*?\<\/div\>","",context,flags=re.DOTALL)
+        context = re.sub(r"\<span\>.*?\<\/span\>","",context,flags=re.DOTALL)
+        if exlist :
+            newtxt = ""
+            for ln in context.splitlines() :
+                matched = False
+                for p in exlist :
+                    if p in ln :
+                        matched = True
+                        break
+                if not matched :
+                    newtxt += ln + "\n"
+            context = newtxt
         if ix == 0 :
             prevpg = "index.html"
             cdone += 1 
@@ -233,12 +289,12 @@ def adhoc_fetch(title,page,limit=10,tgtdir=os.getcwd(),debug=False) :
         rtext += "<br/>\n"*2
         rtext += "</body>\n"
         rtext += "</html>\n"
-        with open(os.path.join(tgtdir,title, cname),"w") as f :
+        with open(os.path.join(tgtdir,title, cname),"w",encoding="UTF-8") as f :
             f.write(rtext)
-        with open(os.path.join(tgtdir,title, "downloaded.txt"),"a") as f :
+        with open(os.path.join(tgtdir,title, "downloaded.txt"),"a",encoding="UTF-8") as f :
             f.write(cname+"\n")
         if cname not in downloaded :
-            with open(os.path.join(tgtdir,title, "index.html"),"a") as f :
+            with open(os.path.join(tgtdir,title, "index.html"),"a",encoding="UTF-8") as f :
                 hcontent = "<tr><td><a href=\"{}\">{}</a></td></tr>".format(cname,chapter) + "\n"
                 f.write(hcontent)
                 if not report :
@@ -252,15 +308,15 @@ def adhoc_fetch(title,page,limit=10,tgtdir=os.getcwd(),debug=False) :
     return report
 
 
-def batch_fetch(cfgfile,num=10,debug=False) :
+def batch_fetch(cfgfile,num=10,debug=False,force=0,exlist=None) :
     report = "\n"
-    with open(cfgfile,"r") as f :
+    with open(cfgfile,"r",encoding="UTF-8") as f :
         thefirst = True
         seconds_wait = 5
         for ln in f.readlines() :
             if not ln or ln.startswith("#") :
                 continue
-            m = re.search(r"(\S+?),(\S+?),(\S+)",ln)
+            m = re.search(r"(\S+?)\,(\S+?)\,(\S+)",ln)
             if m :
                 if thefirst :
                     thefirst = False
@@ -273,9 +329,14 @@ def batch_fetch(cfgfile,num=10,debug=False) :
                 title = m.group(1)
                 tgtdir = m.group(2)
                 page = m.group(3)
+                root = None
+                if "," in page :
+                    arr = page.split(",")
+                    page = arr[0]
+                    root = arr[1]
                 print("#",title,"@", tgtdir, "from", page, file=sys.stderr, flush=True)
                 try :
-                    rpt = adhoc_fetch(title,page,num,tgtdir,debug) + "\n\n"
+                    rpt = adhoc_fetch(title,page,num,tgtdir,debug,force,root,exlist) + "\n"
                     report += rpt
                     if not rpt or not re.search(r"\S+",rpt) or re.search(r"No update",rpt,re.IGNORECASE) :
                         seconds_wait = 5
@@ -288,12 +349,15 @@ def batch_fetch(cfgfile,num=10,debug=False) :
 
 def novels_main():
     parser = argparse.ArgumentParser()
-    parser.add_argument("-f", "--cfgfile", dest="cfgfile", default="~/.novels/novel.conf", help="input file")
+    parser.add_argument("-i", "--cfgfile", dest="cfgfile", default="~/.novels/novel.conf", help="input file")
     parser.add_argument("-t", "--title", dest="title", default=None, help="novel title")
     parser.add_argument("-n", "--number", dest="number", type=int, default=10, help="how many chapters to retrieve")
     parser.add_argument("-p", "--page", dest="page", default=None, help="Novel index page",)
     parser.add_argument("-d", "--dir", dest="dir", default=os.getcwd(), help="where to store books",)
+    parser.add_argument("-E", "--excludes", dest="excludes", default="~/.novels/excludes.conf", help="pattern or file of patterns to be excluded",)
+    parser.add_argument("-r", "--root", dest="root", default=None, help="page root. use it when index page has differnt root then chapters.",)
     parser.add_argument("-I", "--bookindex", dest="bookindex", action="store_true",default=False, help="build index page for latest chapters.",)
+    parser.add_argument("-f", "--force", dest="force", action="count", default=0, help="force reload last chapter when no update.",)
     parser.add_argument("-X", "--debug", dest="debug", action="count", default=0, help="debug mode",)
     args = parser.parse_args()
 
@@ -301,18 +365,23 @@ def novels_main():
         build_index(args.dir) 
         return
 
+    exlist = None
+    if args.excludes :
+        args.excludes = os.path.expanduser(args.excludes)
+        exlist = get_exlist(args.excludes)
+
     if args.title :
-        adhoc_fetch(args.title,args.page,args.number,args.dir,args.debug)
+        adhoc_fetch(args.title,args.page,args.number,args.dir,args.debug,args.force,args.root,exlist)
         return
     args.cfgfile = os.path.expanduser(args.cfgfile)
     if os.path.isfile(args.cfgfile) :
-        report = batch_fetch(args.cfgfile,num=args.number,debug=args.debug)
+        report = batch_fetch(args.cfgfile,num=args.number,debug=args.debug,force=args.force,exlist=exlist)
         print("# Summary :\n" + report)
     else :
         if not os.path.isdir(os.path.expanduser("~/.novels")) :
             os.mkdir(os.path.expanduser("~/.novels"))
-        with open(args.cfgfile,"w") as f :
-            f.write("#title,bookpage\n")
+        with open(args.cfgfile,"w",encoding="UTF-8") as f :
+            f.write("#title,localdir,bookpage\n")
 
 
 if __name__ == "__main__" :
